#include "types.h"
#include "param.h"
#include "memlayout.h"
#include "riscv.h"
#include "spinlock.h"
#include "proc.h"
#include "defs.h"
#include "rl.h"  // 包含强化学习相关的定义

// 其他代码

struct cpu cpus[NCPU];

struct proc proc[NPROC];

struct proc *initproc;

int nextpid = 1;
struct spinlock pid_lock;

extern void forkret(void);
static void wakeup1(struct proc *chan);
static void freeproc(struct proc *p);

extern char trampoline[]; // trampoline.S

static unsigned int seed = 10000007; // 存储随机数生成器的种子。  
// proc.c

struct complex_policy policy;          // 策略网络
struct replay_buffer buffer;           // 经验回放缓冲区
int policy_initialized = 0;            // 策略网络初始化标志
int buffer_initialized = 0;            // 经验回放缓冲区初始化标志
void init_replay_buffer(struct replay_buffer *buffer) {
    buffer->size = 0;
    buffer->head = 0;
}

void add_experience(struct replay_buffer *buffer, struct rl_state state, struct rl_action action, int reward, struct rl_state next_state) {
    buffer->buffer[buffer->head] = (struct experience){state, action, reward, next_state};
    buffer->head = (buffer->head + 1) % REPLAY_BUFFER_SIZE;
    if (buffer->size < REPLAY_BUFFER_SIZE) {
        buffer->size++;
    }
}

struct experience sample_experience(struct replay_buffer *buffer) {
    int index = random_at_most(buffer->size - 1);
    return buffer->buffer[index];
}

int calculate_reward(int vruntime, int tickets, int total_tickets) {
    int reward = 100 - vruntime;  // 奖励函数（基于 vruntime）
    reward -= total_tickets / 10;  // 惩罚高负载
    return reward;
}

struct rl_action epsilon_greedy(struct rl_state state, struct complex_policy *policy) {
    if (random_at_most(100) < EPSILON * 100) {
        // 随机选择动作
        struct rl_action random_action;
        random_action.delta_tickets = random_at_most(20) - 10;  // 随机调整彩票数
        random_action.next_pid = random_at_most(NPROC);  // 随机选择下一个进程
        return random_action;
    } else {
        // 使用策略网络选择动作
        return choose_action_complex(state, policy);
    }
}
// 初始化策略网络
void init_complex_policy(struct complex_policy *policy) {
    for (int i = 0; i < STATE_DIM; i++) {
        for (int j = 0; j < HIDDEN_DIM; j++) {
            policy->weights1[i][j] = random_at_most(100);  // 随机初始化权重
        }
    }
    for (int i = 0; i < HIDDEN_DIM; i++) {
        for (int j = 0; j < ACTION_DIM; j++) {
            policy->weights2[i][j] = random_at_most(100);  // 随机初始化权重
        }
    }
    for (int i = 0; i < HIDDEN_DIM; i++) {
        policy->bias1[i] = random_at_most(100);  // 随机初始化偏置
    }
    for (int i = 0; i < ACTION_DIM; i++) {
        policy->bias2[i] = random_at_most(100);  // 随机初始化偏置
    }
}

// 根据状态选择动作（简化版）
struct rl_action choose_action_complex(struct rl_state state, struct complex_policy *policy) {
    struct rl_action action;
    int state_vector[STATE_DIM] = {
        state.vruntime,
        state.tickets,
        state.cpu_load,
        state.elapsed_time
    };

    // 第一层计算
    int hidden[HIDDEN_DIM] = {0};
    for (int i = 0; i < HIDDEN_DIM; i++) {
        for (int j = 0; j < STATE_DIM; j++) {
            hidden[i] += (state_vector[j] * policy->weights1[j][i]) / 100;  // 整数运算
        }
        hidden[i] += policy->bias1[i];  // 加上偏置
        hidden[i] = hidden[i] > 0 ? hidden[i] : 0;  // ReLU激活函数
    }

    // 第二层计算
    action.delta_tickets = 0;
    action.next_pid = 0;
    for (int i = 0; i < ACTION_DIM; i++) {
        for (int j = 0; j < HIDDEN_DIM; j++) {
            if (i == 0) {
                action.delta_tickets += (hidden[j] * policy->weights2[j][i]) / 100;  // 整数运算
            } else {
                action.next_pid += (hidden[j] * policy->weights2[j][i]) / 100;  // 整数运算
            }
        }
        if (i == 0) {
            action.delta_tickets += policy->bias2[i];  // 加上偏置
        } else {
            action.next_pid += policy->bias2[i];  // 加上偏置
        }
    }

    return action;
}

// 更新策略网络（简化版）
void update_complex_policy(struct complex_policy *policy, int error) {
    int learning_rate = 1;  // 学习率（整数）
    for (int i = 0; i < STATE_DIM; i++) {
        for (int j = 0; j < HIDDEN_DIM; j++) {
            policy->weights1[i][j] += (learning_rate * error) / 100;  // 使用学习率更新权重
            if (policy->weights1[i][j] > 100) policy->weights1[i][j] = 100;
            if (policy->weights1[i][j] < 0) policy->weights1[i][j] = 0;
        }
    }
    for (int i = 0; i < HIDDEN_DIM; i++) {
        for (int j = 0; j < ACTION_DIM; j++) {
            policy->weights2[i][j] += (learning_rate * error) / 100;  // 使用学习率更新权重
            if (policy->weights2[i][j] > 100) policy->weights2[i][j] = 100;
            if (policy->weights2[i][j] < 0) policy->weights2[i][j] = 0;
        }
    }
}
int
random(void)
{
  seed = (1103515245 * seed + 12345) % 2147483648;  // 线性同余生成器LCG算法生成伪随机数。
  return seed; // 返回伪随机数
}

int
random_at_most(int max) // 生成一个在0到max之间的随机数。
{
  unsigned long num_bins = (unsigned long) max + 1; // 目标区间的大小0~max即为max+1
  unsigned long num_rand = (unsigned long) 2147483647 + 1; // 随机数生成器的范围
  unsigned long bin_size = num_rand / num_bins; // 每个区间的平均大小
  unsigned long defect = num_rand % num_bins; // 剩余的无法均匀分配的部分
  int x;
  do {
    x = random();
  } while (num_rand - defect <= (unsigned long)x); //确保生成的随机数x不会落在无法均匀分配的部分。
  return x / bin_size; // 将随机数映射到目标区间0~max
}
static int total_tickets = 0;  // 用于存储所有进程的总彩票数

static void
init_lottery(struct proc *p)
{
  p->tickets = 10; // 初始化进程的彩票数
  p->old_tickets = p->tickets; // 保存当前彩票数，用于后续更新
  p->vruntime = 0; // 初始化进程的虚拟运行时间
  p->elapsed_time = 0; // 初始化进程的已运行时间
  p->cpu = -1; // 初始化进程的 CPU 绑定状态
  total_tickets += p->tickets;
}

void update_tickets(struct proc *p) {
    int old_tickets = p->tickets;
    struct rl_state state = {
        .vruntime = p->vruntime,
        .tickets = p->tickets,
        .cpu_load = total_tickets,  // 用总彩票数表示CPU负载
        .elapsed_time = p->elapsed_time
    };

    // 使用复杂版策略网络选择动作
    struct rl_action action = epsilon_greedy(state, &policy);
    p->tickets += action.delta_tickets;

    // 限制彩票数量的范围
    if (p->tickets < 5) p->tickets = 5;
    if (p->tickets > 50) p->tickets = 50;

    // 更新总彩票数，并防止 total_tickets 为负数
    total_tickets += p->tickets - old_tickets;
    if (total_tickets < 0) {
        total_tickets = 0;
    }
}



// initialize the proc table at boot time.
void procinit(void) {
    struct proc *p;
    initlock(&pid_lock, "nextpid");

    // 初始化简化版策略网络
    init_complex_policy(&policy);

    for (p = proc; p < &proc[NPROC]; p++) {
        initlock(&p->lock, "proc");
        char *pa = kalloc();
        if (pa == 0)
            panic("kalloc");
        uint64 va = KSTACK((int)(p - proc));
        kvmmap(va, (uint64)pa, PGSIZE, PTE_R | PTE_W);
        p->kstack = va;
    }
    kvminithart();
}
// Must be called with interrupts disabled,
// to prevent race with process being moved
// to a different CPU.
int
cpuid()
{
  int id = r_tp();
  return id;
}

// Return this CPU's cpu struct.
// Interrupts must be disabled.
struct cpu*
mycpu(void) {
  int id = cpuid();
  struct cpu *c = &cpus[id];
  return c;
}

// Return the current struct proc *, or zero if none.
struct proc*
myproc(void) {
  push_off();
  struct cpu *c = mycpu();
  struct proc *p = c->proc;
  pop_off();
  return p;
}

int
allocpid() {
  int pid;
  
  acquire(&pid_lock);
  pid = nextpid;
  nextpid = nextpid + 1;
  release(&pid_lock);

  return pid;
}

// Look in the process table for an UNUSED proc.
// If found, initialize state required to run in the kernel,
// and return with p->lock held.
// If there are no free procs, or a memory allocation fails, return 0.
/*
static struct proc*
allocproc(void)
{
  struct proc *p;

  for(p = proc; p < &proc[NPROC]; p++) {
    acquire(&p->lock);
    if(p->state == UNUSED) {
      goto found;
    } else {
      release(&p->lock);
    }
  }
  return 0;

found:
  p->pid = allocpid();

  // Allocate a trapframe page.
  if((p->trapframe = (struct trapframe *)kalloc()) == 0){
    release(&p->lock);
    return 0;
  }

  // An empty user page table.
  p->pagetable = proc_pagetable(p);
  if(p->pagetable == 0){
    freeproc(p);
    release(&p->lock);
    return 0;
  }

  // Set up new context to start executing at forkret,
  // which returns to user space.
  memset(&p->context, 0, sizeof(p->context));
  p->context.ra = (uint64)forkret;
  p->context.sp = p->kstack + PGSIZE;

  return p;
}
*/
static struct proc*
allocproc(void)
{
  struct proc *p;

  for(p = proc; p < &proc[NPROC]; p++) {
    acquire(&p->lock);
    if(p->state == UNUSED) {
      goto found;
    } else {
      release(&p->lock);
    }
  }
  return 0;

found:
  p->pid = allocpid();

  // Allocate a trapframe page.
  if((p->trapframe = (struct trapframe *)kalloc()) == 0){
    release(&p->lock);
    return 0;
  }

  // An empty user page table.
  p->pagetable = proc_pagetable(p);
  if(p->pagetable == 0){
    freeproc(p);
    release(&p->lock);
    return 0;
  }

  // Set up new context to start executing at forkret,
  // which returns to user space.
  memset(&p->context, 0, sizeof(p->context));
  p->context.ra = (uint64)forkret;
  p->context.sp = p->kstack + PGSIZE;

  init_lottery(p); //初始化

  return p;
}
// free a proc structure and the data hanging from it,
// including user pages.
// p->lock must be held.
static void
freeproc(struct proc *p)
{
  if(p->trapframe)
    kfree((void*)p->trapframe);
  p->trapframe = 0;
  if(p->pagetable)
    proc_freepagetable(p->pagetable, p->sz);
  p->pagetable = 0;
  p->sz = 0;
  p->pid = 0;
  p->parent = 0;
  p->name[0] = 0;
  p->chan = 0;
  p->killed = 0;
  p->xstate = 0;
  p->state = UNUSED;
}

// Create a user page table for a given process,
// with no user memory, but with trampoline pages.
pagetable_t
proc_pagetable(struct proc *p)
{
  pagetable_t pagetable;

  // An empty page table.
  pagetable = uvmcreate();
  if(pagetable == 0)
    return 0;

  // map the trampoline code (for system call return)
  // at the highest user virtual address.
  // only the supervisor uses it, on the way
  // to/from user space, so not PTE_U.
  if(mappages(pagetable, TRAMPOLINE, PGSIZE,
              (uint64)trampoline, PTE_R | PTE_X) < 0){
    uvmfree(pagetable, 0);
    return 0;
  }

  // map the trapframe just below TRAMPOLINE, for trampoline.S.
  if(mappages(pagetable, TRAPFRAME, PGSIZE,
              (uint64)(p->trapframe), PTE_R | PTE_W) < 0){
    uvmunmap(pagetable, TRAMPOLINE, 1, 0);
    uvmfree(pagetable, 0);
    return 0;
  }

  return pagetable;
}

// Free a process's page table, and free the
// physical memory it refers to.
void
proc_freepagetable(pagetable_t pagetable, uint64 sz)
{
  uvmunmap(pagetable, TRAMPOLINE, 1, 0);
  uvmunmap(pagetable, TRAPFRAME, 1, 0);
  uvmfree(pagetable, sz);
}

// a user program that calls exec("/init")
// od -t xC initcode
uchar initcode[] = {
  0x17, 0x05, 0x00, 0x00, 0x13, 0x05, 0x45, 0x02,
  0x97, 0x05, 0x00, 0x00, 0x93, 0x85, 0x35, 0x02,
  0x93, 0x08, 0x70, 0x00, 0x73, 0x00, 0x00, 0x00,
  0x93, 0x08, 0x20, 0x00, 0x73, 0x00, 0x00, 0x00,
  0xef, 0xf0, 0x9f, 0xff, 0x2f, 0x69, 0x6e, 0x69,
  0x74, 0x00, 0x00, 0x24, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00
};

// Set up first user process.
void
userinit(void)
{
  struct proc *p;

  p = allocproc();
  initproc = p;
  
  // allocate one user page and copy init's instructions
  // and data into it.
  uvminit(p->pagetable, initcode, sizeof(initcode));
  p->sz = PGSIZE;

  // prepare for the very first "return" from kernel to user.
  p->trapframe->epc = 0;      // user program counter
  p->trapframe->sp = PGSIZE;  // user stack pointer

  safestrcpy(p->name, "initcode", sizeof(p->name));
  p->cwd = namei("/");

  p->state = RUNNABLE;

  release(&p->lock);
}

// Grow or shrink user memory by n bytes.
// Return 0 on success, -1 on failure.
int
growproc(int n)
{
  uint sz;
  struct proc *p = myproc();

  sz = p->sz;
  if(n > 0){
    if((sz = uvmalloc(p->pagetable, sz, sz + n)) == 0) {
      return -1;
    }
  } else if(n < 0){
    sz = uvmdealloc(p->pagetable, sz, sz + n);
  }
  p->sz = sz;
  return 0;
}

// Create a new process, copying the parent.
// Sets up child kernel stack to return as if from fork() system call.
int
fork(void)
{
  int i, pid;
  struct proc *np;
  struct proc *p = myproc();

  // Allocate process.
  if((np = allocproc()) == 0){
    return -1;
  }

  // Copy user memory from parent to child.
  if(uvmcopy(p->pagetable, np->pagetable, p->sz) < 0){
    freeproc(np);
    release(&np->lock);
    return -1;
  }
  np->sz = p->sz;

  np->parent = p;

  // copy saved user registers.
  *(np->trapframe) = *(p->trapframe);

  // Cause fork to return 0 in the child.
  np->trapframe->a0 = 0;

  // increment reference counts on open file descriptors.
  for(i = 0; i < NOFILE; i++)
    if(p->ofile[i])
      np->ofile[i] = filedup(p->ofile[i]);
  np->cwd = idup(p->cwd);

  safestrcpy(np->name, p->name, sizeof(p->name));

  pid = np->pid;

  np->state = RUNNABLE;

  release(&np->lock);

  return pid;
}

// Pass p's abandoned children to init.
// Caller must hold p->lock.
void
reparent(struct proc *p)
{
  struct proc *pp;

  for(pp = proc; pp < &proc[NPROC]; pp++){
    // this code uses pp->parent without holding pp->lock.
    // acquiring the lock first could cause a deadlock
    // if pp or a child of pp were also in exit()
    // and about to try to lock p.
    if(pp->parent == p){
      // pp->parent can't change between the check and the acquire()
      // because only the parent changes it, and we're the parent.
      acquire(&pp->lock);
      pp->parent = initproc;
      // we should wake up init here, but that would require
      // initproc->lock, which would be a deadlock, since we hold
      // the lock on one of init's children (pp). this is why
      // exit() always wakes init (before acquiring any locks).
      release(&pp->lock);
    }
  }
}

// Exit the current process.  Does not return.
// An exited process remains in the zombie state
// until its parent calls wait().
void
exit(int status)
{
  struct proc *p = myproc();

  if(p == initproc)
    panic("init exiting");

  // Close all open files.
  for(int fd = 0; fd < NOFILE; fd++){
    if(p->ofile[fd]){
      struct file *f = p->ofile[fd];
      fileclose(f);
      p->ofile[fd] = 0;
    }
  }

  begin_op();
  iput(p->cwd);
  end_op();
  p->cwd = 0;

  // we might re-parent a child to init. we can't be precise about
  // waking up init, since we can't acquire its lock once we've
  // acquired any other proc lock. so wake up init whether that's
  // necessary or not. init may miss this wakeup, but that seems
  // harmless.
  acquire(&initproc->lock);
  wakeup1(initproc);
  release(&initproc->lock);

  // grab a copy of p->parent, to ensure that we unlock the same
  // parent we locked. in case our parent gives us away to init while
  // we're waiting for the parent lock. we may then race with an
  // exiting parent, but the result will be a harmless spurious wakeup
  // to a dead or wrong process; proc structs are never re-allocated
  // as anything else.
  acquire(&p->lock);
  struct proc *original_parent = p->parent;
  release(&p->lock);
  
  // we need the parent's lock in order to wake it up from wait().
  // the parent-then-child rule says we have to lock it first.
  acquire(&original_parent->lock);

  acquire(&p->lock);

  // Give any children to init.
  reparent(p);

  // Parent might be sleeping in wait().
  wakeup1(original_parent);

  p->xstate = status;
  p->state = ZOMBIE;

  release(&original_parent->lock);

  // Jump into the scheduler, never to return.
  sched();
  panic("zombie exit");
}

// Wait for a child process to exit and return its pid.
// Return -1 if this process has no children.
int
wait(uint64 addr)
{
  struct proc *np;
  int havekids, pid;
  struct proc *p = myproc();

  // hold p->lock for the whole time to avoid lost
  // wakeups from a child's exit().
  acquire(&p->lock);

  for(;;){
    // Scan through table looking for exited children.
    havekids = 0;
    for(np = proc; np < &proc[NPROC]; np++){
      // this code uses np->parent without holding np->lock.
      // acquiring the lock first would cause a deadlock,
      // since np might be an ancestor, and we already hold p->lock.
      if(np->parent == p){
        // np->parent can't change between the check and the acquire()
        // because only the parent changes it, and we're the parent.
        acquire(&np->lock);
        havekids = 1;
        if(np->state == ZOMBIE){
          // Found one.
          pid = np->pid;
          if(addr != 0 && copyout(p->pagetable, addr, (char *)&np->xstate,
                                  sizeof(np->xstate)) < 0) {
            release(&np->lock);
            release(&p->lock);
            return -1;
          }
          freeproc(np);
          release(&np->lock);
          release(&p->lock);
          return pid;
        }
        release(&np->lock);
      }
    }

    // No point waiting if we don't have any children.
    if(!havekids || p->killed){
      release(&p->lock);
      return -1;
    }
    
    // Wait for a child to exit.
    sleep(p, &p->lock);  //DOC: wait-sleep
  }
}

static struct proc* pick_process(void) {
    if (total_tickets == 0) {
        return 0;  // 没有可运行进程
    }
    int winner = random_at_most(total_tickets - 1);  // 随机选择一个彩票赢家
    int counter = 0; // 初始化彩票计数器为0
    int current_cpu_id = cpuid();  // 获取当前CPU的ID

    // 优先考虑绑定到当前CPU的进程
    for (struct proc *p = proc; p < &proc[NPROC]; p++) {
        if (p->state == RUNNABLE && p->cpu == current_cpu_id) { // 检查进程是否是RUNNABLE状态，并且是否绑定到当前CPU
            counter += p->tickets;  // 累加当前进程的彩票数到counter
            if (counter > winner) { // 如果counter超过winner，说明当前进程是中奖进程
                return p;  // 返回中奖进程
            }
        }
    }

    // 如果没有找到绑定到当前CPU的进程，考虑其他CPU上的进程
    for (struct proc *p = proc; p < &proc[NPROC]; p++) {
        if (p->state == RUNNABLE) { // 与上一种情况类似同理
            counter += p->tickets;  
            if (counter > winner) {
                return p;  
            }
        }
    }
    return 0;  
}
int predict_reward(struct complex_policy *policy, struct rl_state state) {
    int state_vector[STATE_DIM] = {
        state.vruntime,
        state.tickets,
        state.cpu_load,
        state.elapsed_time
    };

    // 第一层计算（隐藏层）
    int hidden[HIDDEN_DIM] = {0};
    for (int i = 0; i < HIDDEN_DIM; i++) {
        for (int j = 0; j < STATE_DIM; j++) {
            hidden[i] += (state_vector[j] * policy->weights1[j][i]) / 100;  // 整数运算
        }
        hidden[i] += policy->bias1[i];  // 加上偏置
        hidden[i] = hidden[i] > 0 ? hidden[i] : 0;  // ReLU激活函数
    }

    // 第二层计算（输出层）
    int predicted_value = 0;
    for (int i = 0; i < ACTION_DIM; i++) {
        for (int j = 0; j < HIDDEN_DIM; j++) {
            predicted_value += (hidden[j] * policy->weights2[j][i]) / 100;  // 整数运算
        }
        predicted_value += policy->bias2[i];  // 加上偏置
    }

    return predicted_value;
}
int calculate_target_value(int reward, struct rl_state next_state, struct complex_policy *policy) {
    int gamma = 9;  // 折扣因子（0.9 * 10，使用整数表示）
    int future_reward = predict_reward(policy, next_state);  // 预测未来奖励
    int target_value = reward * 10 + gamma * future_reward;  // 计算目标值（放大 10 倍）
    return target_value / 10;  // 缩小 10 倍
}

int calculate_error(int reward, struct rl_state state, struct rl_state next_state, struct complex_policy *policy) {
    // 计算目标值
    int target_value = calculate_target_value(reward, next_state, policy);
    // 计算预测值
    int predicted_value = predict_reward(policy, state);
    // 计算误差
    int error = target_value - predicted_value;
    return error;
}
// Per-CPU process scheduler.
// Each CPU calls scheduler() after setting itself up.
// Scheduler never returns.  It loops, doing:
//  - choose a process to run.
//  - swtch to start running that process.
//  - eventually that process transfers control
//    via swtch back to the scheduler.

void scheduler(void) {
    struct proc *p;
    struct cpu *c = mycpu();
    c->proc = 0;

    // 初始化策略网络和经验回放缓冲区
    static struct complex_policy policy;
    static int policy_initialized = 0;
    if (!policy_initialized) {
        init_complex_policy(&policy);
        policy_initialized = 1;
    }

    static struct replay_buffer buffer;
    static int buffer_initialized = 0;
    if (!buffer_initialized) {
        init_replay_buffer(&buffer);
        buffer_initialized = 1;
    }

    for (;;) {
        intr_on();

        // 更新所有可运行进程的票数
        for (p = proc; p < &proc[NPROC]; p++) {
            if (p->state == RUNNABLE) {
                update_tickets(p);
            }
        }

        // 选择下一个运行的进程
        p = pick_process();
        if (p) {
            acquire(&p->lock);
            if (p->state == RUNNABLE) {
                p->state = RUNNING;
                p->cpu = cpuid();
                c->proc = p;
                swtch(&c->context, &p->context);

                // 更新虚拟运行时间和奖励
                p->vruntime += p->elapsed_time / (p->tickets + 1);
                int reward = calculate_reward(p->vruntime, p->tickets, total_tickets);

                // 计算误差
                struct rl_state state = {
                    .vruntime = p->vruntime,
                    .tickets = p->tickets,
                    .cpu_load = total_tickets,
                    .elapsed_time = p->elapsed_time
                };
                struct rl_state next_state = {
                    .vruntime = p->vruntime + p->elapsed_time / (p->tickets + 1),
                    .tickets = p->tickets,
                    .cpu_load = total_tickets,
                    .elapsed_time = 0
                };
                int error = calculate_error(reward, state, next_state, &policy);

                // 更新策略网络
                update_complex_policy(&policy, error);

                p->elapsed_time = 0;
                c->proc = 0;
            }
            release(&p->lock);
        } else {
            intr_on();
            asm volatile("wfi");
        }
    }
}
// Switch to scheduler.  Must hold only p->lock
// and have changed proc->state. Saves and restores
// intena because intena is a property of this
// kernel thread, not this CPU. It should
// be proc->intena and proc->noff, but that would
// break in the few places where a lock is held but
// there's no process.
void
sched(void)
{
  int intena;
  struct proc *p = myproc();

  if(!holding(&p->lock))
    panic("sched p->lock");
  if(mycpu()->noff != 1)
    panic("sched locks");
  if(p->state == RUNNING)
    panic("sched running");
  if(intr_get())
    panic("sched interruptible");

  intena = mycpu()->intena;
  swtch(&p->context, &mycpu()->context);
  mycpu()->intena = intena;
}

// Give up the CPU for one scheduling round.
void
yield(void)
{
  struct proc *p = myproc();
  acquire(&p->lock);
  p->state = RUNNABLE;
  sched();
  release(&p->lock);
}

// A fork child's very first scheduling by scheduler()
// will swtch to forkret.
void
forkret(void)
{
  static int first = 1;

  // Still holding p->lock from scheduler.
  release(&myproc()->lock);

  if (first) {
    // File system initialization must be run in the context of a
    // regular process (e.g., because it calls sleep), and thus cannot
    // be run from main().
    first = 0;
    fsinit(ROOTDEV);
  }

  usertrapret();
}

// Atomically release lock and sleep on chan.
// Reacquires lock when awakened.
void
sleep(void *chan, struct spinlock *lk)
{
  struct proc *p = myproc();
  
  // Must acquire p->lock in order to
  // change p->state and then call sched.
  // Once we hold p->lock, we can be
  // guaranteed that we won't miss any wakeup
  // (wakeup locks p->lock),
  // so it's okay to release lk.
  if(lk != &p->lock){  //DOC: sleeplock0
    acquire(&p->lock);  //DOC: sleeplock1
    release(lk);
  }

  // Go to sleep.
  p->chan = chan;
  p->state = SLEEPING;

  sched();

  // Tidy up.
  p->chan = 0;

  // Reacquire original lock.
  if(lk != &p->lock){
    release(&p->lock);
    acquire(lk);
  }
}

// Wake up all processes sleeping on chan.
// Must be called without any p->lock.
void
wakeup(void *chan)
{
  struct proc *p;

  for(p = proc; p < &proc[NPROC]; p++) {
    acquire(&p->lock);
    if(p->state == SLEEPING && p->chan == chan) {
      p->state = RUNNABLE;
    }
    release(&p->lock);
  }
}

// Wake up p if it is sleeping in wait(); used by exit().
// Caller must hold p->lock.
static void
wakeup1(struct proc *p)
{
  if(!holding(&p->lock))
    panic("wakeup1");
  if(p->chan == p && p->state == SLEEPING) {
    p->state = RUNNABLE;
  }
}

// Kill the process with the given pid.
// The victim won't exit until it tries to return
// to user space (see usertrap() in trap.c).
int
kill(int pid)
{
  struct proc *p;

  for(p = proc; p < &proc[NPROC]; p++){
    acquire(&p->lock);
    if(p->pid == pid){
      p->killed = 1;
      if(p->state == SLEEPING){
        // Wake process from sleep().
        p->state = RUNNABLE;
      }
      release(&p->lock);
      return 0;
    }
    release(&p->lock);
  }
  return -1;
}

// Copy to either a user address, or kernel address,
// depending on usr_dst.
// Returns 0 on success, -1 on error.
int
either_copyout(int user_dst, uint64 dst, void *src, uint64 len)
{
  struct proc *p = myproc();
  if(user_dst){
    return copyout(p->pagetable, dst, src, len);
  } else {
    memmove((char *)dst, src, len);
    return 0;
  }
}

// Copy from either a user address, or kernel address,
// depending on usr_src.
// Returns 0 on success, -1 on error.
int
either_copyin(void *dst, int user_src, uint64 src, uint64 len)
{
  struct proc *p = myproc();
  if(user_src){
    return copyin(p->pagetable, dst, src, len);
  } else {
    memmove(dst, (char*)src, len);
    return 0;
  }
}

// Print a process listing to console.  For debugging.
// Runs when user types ^P on console.
// No lock to avoid wedging a stuck machine further.
void
procdump(void)
{
  static char *states[] = {
  [UNUSED]    "unused",
  [SLEEPING]  "sleep ",
  [RUNNABLE]  "runble",
  [RUNNING]   "run   ",
  [ZOMBIE]    "zombie"
  };
  struct proc *p;
  char *state;

  printf("\n");
  for(p = proc; p < &proc[NPROC]; p++){
    if(p->state == UNUSED)
      continue;
    if(p->state >= 0 && p->state < NELEM(states) && states[p->state])
      state = states[p->state];
    else
      state = "???";
    printf("%d %s %s", p->pid, state, p->name);
    printf("\n");
  }
}



