#include "kernel/types.h"
#include "kernel/stat.h"
#include "user/user.h"

// 计算斐波那契数列的第 n 项
int fib(int n) {
    if (n <= 1) return n;
    return fib(n - 1) + fib(n - 2);
}

// 计算密集型任务采用计算斐波那契数列
void compute_task(int id) {
    int start_time = uptime();
    int result = 0;
    for (int i = 0; i < 37; i++) {  
        result += fib(i);
    }
    int end_time = uptime();
    printf("Compute Task %d finished in %d ticks\n", id, end_time - start_time);
    exit((int)(end_time - start_time)); 
}

// I/O 密集型任务：模拟 I/O 操作
void io_task(int id) {
    int start_time = uptime();
    for (int i = 0; i < 5; i++) {
        sleep(5); // 模拟 I/O 等待
    }
    int end_time = uptime();
    printf("I/O Task %d finished in %d ticks\n", id, end_time - start_time);
    exit((int)(end_time - start_time));
}

// 混合任务：计算密集型 + I/O 密集型
void mixed_task(int id) {
    int start_time = uptime();
    for (int i = 0; i < 3; i++) {
        // 计算密集型操作
        int compute_start = uptime();
        int result = 0;
        for (int j = 0; j < 35; j++) {  
            result += fib(j);
        }
        int compute_end = uptime();

        // I/O 密集型操作
        int io_start = uptime();
        sleep(5); // 模拟 I/O 等待
        int io_end = uptime();

        printf("Mixed Task %d - Iteration %d: Compute %d ticks, I/O %d ticks\n",
            id, i, compute_end - compute_start, io_end - io_start);
    }
    int end_time = uptime();
    printf("Mixed Task %d finished in %d ticks\n", id, end_time - start_time);
    exit((int)(end_time - start_time)); 
}

int main(int argc, char *argv[]) {
    int num_compute = 2; // 计算密集型进程数量
    int num_io = 2;      // I/O 密集型进程数量
    int num_mixed = 2;   // 混合型进程数量

    // 用于存储运行时间的数组
    int compute_times[num_compute];
    int io_times[num_io];
    int mixed_times[num_mixed];

    // 用于存储等待时间的数组
    int compute_waits[num_compute];
    int io_waits[num_io];
    int mixed_waits[num_mixed];

    // 定义子进程信息
    struct {
        int pid;
        int type;
        int start_time; // 记录任务的开始时间
    } children[num_compute + num_io + num_mixed + 1];

    int index = 0;

    // 创建计算密集型进程
    for (int i = 0; i < num_compute; i++) {
        int pid = fork();
        if (pid == 0) {
            compute_task(i);
            exit(0);
        } else {
            children[index].pid = pid;
            children[index].type = 0; // 0表示计算任务
            children[index].start_time = uptime();
            index++;
        }
    }

    // 创建 I/O 密集型进程
    for (int i = 0; i < num_io; i++) {
        int pid = fork();
        if (pid == 0) {
            io_task(i + num_compute);
            exit(0);
        } else {
            children[index].pid = pid;
            children[index].type = 1; // 1表示I/O任务
            children[index].start_time = uptime(); 
            index++;
        }
    }

    // 创建混合型进程
    for (int i = 0; i < num_mixed; i++) {
        int pid = fork();
        if (pid == 0) {
            mixed_task(i + num_compute + num_io);
            exit(0);
        } else {
            children[index].pid = pid;
            children[index].type = 2; // 2表示混合任务
            children[index].start_time = uptime(); 
            index++;
        }
    }

    // 等待所有子进程完成并收集运行时间和等待时间
    int compute_index = 0;
    int io_index = 0;
    int mixed_index = 0;

    for (int i = 0; i < num_compute + num_io + num_mixed; i++) {
        int status;
        int pid = wait(&status);
        if (pid < 0) {
            break;
        }
        int runtime = status; // 提取子进程的退出码
        int end_time = uptime(); 

        // 查找这个pid对应的type
        for (int j = 0; j < num_compute + num_io + num_mixed; j++) {
            if (children[j].pid == pid) {
                int wait_time = end_time - children[j].start_time;
                switch (children[j].type) {
                    case 0:
                        compute_times[compute_index] = runtime;
                        compute_waits[compute_index] = wait_time;
                        compute_index++;
                        break;
                    case 1:
                        io_times[io_index] = runtime;
                        io_waits[io_index] = wait_time;
                        io_index++;
                        break;
                    case 2:
                        mixed_times[mixed_index] = runtime;
                        mixed_waits[mixed_index] = wait_time;
                        mixed_index++;
                        break;
                }
                break;
            }
        }
    }

    // 计算总结数据
    int total_compute = 0, total_io = 0, total_mixed = 0;
    int total_compute_wait = 0, total_io_wait = 0, total_mixed_wait = 0;
    for (int i = 0; i < num_compute; i++) {
        total_compute += compute_times[i];
        total_compute_wait += compute_waits[i];
    }
    for (int i = 0; i < num_io; i++) {
        total_io += io_times[i];
        total_io_wait += io_waits[i];
    }
    for (int i = 0; i < num_mixed; i++) {
        total_mixed += mixed_times[i];
        total_mixed_wait += mixed_waits[i];
    }

    int total_time = total_compute + total_io + total_mixed;
    
    int avg_compute = total_compute / num_compute;
    int avg_io = total_io / num_io;
    int avg_mixed = total_mixed / num_mixed;

    int avg_compute_wait = total_compute_wait / num_compute;
    int avg_io_wait = total_io_wait / num_io;
    int avg_mixed_wait = total_mixed_wait / num_mixed;

    // 输出总结数据
    printf("\nSummary:\n");
    printf("Total Compute Tasks: %d, Total Time: %d ticks, Average Time: %d ticks, Average Wait: %d ticks\n", 
           num_compute, total_compute, avg_compute, avg_compute_wait);
    printf("Total I/O Tasks: %d, Total Time: %d ticks, Average Time: %d ticks, Average Wait: %d ticks\n", 
           num_io, total_io, avg_io, avg_io_wait);
    printf("Total Mixed Tasks: %d, Total Time: %d ticks, Average Time: %d ticks, Average Wait: %d ticks\n", 
           num_mixed, total_mixed, avg_mixed, avg_mixed_wait);
    printf("Total Tasks: %d, Total Time: %d ticks\n", num_compute + num_io + num_mixed, total_time);

    // 计算系统吞吐量
    int total_tasks = num_compute + num_io + num_mixed;
    int throughput = total_tasks / (total_time / 100); // 原时间单位为tick，转换为秒
    printf("System Throughput: %d tasks/second\n", throughput);

    exit(0);
}